from pydantic import BaseModel


class ModelInput(BaseModel):
    prompt: str


class ModelOutput(BaseModel):
    generated_text: str


class Model:
    def predict(self, model_input: ModelInput) -> ModelOutput:
        return ModelOutput(generated_text=model_input.prompt)



import time
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest
import requests

from truss.base.truss_config import ModelCache, ModelRepo
from truss.util.basetenpointer import model_cache_hf_to_b10ptr


def test_dolly_12b():
    ModelCached = ModelCache(
        [
            dict(
                repo_id="databricks/dolly-v2-12b",
                revision="19308160448536e378e3db21a73a751579ee7fdd",
                use_volume=True,
                volume_folder="databricks_dolly_v2_12b",
            )
        ]
    )
    for _ in range(2):
        try:
            bptr = model_cache_hf_to_b10ptr(ModelCached)
            continue
        # timeout by huggingface hub timeout error
        except requests.exceptions.ReadTimeout as e:
            # this is expected to timeout when the request takes too long
            # due to the large size of the model
            print("ReadTimeout Error: ", e)
            pytest.skip(
                "Skipping test due to ReadTimeout error from Hugging Face API, "
                "this can happen for large models like Dolly-12b"
            )
    bptr_list = bptr.pointers
    expected = [
        {
            "resolution": {
                "url": "https://huggingface.co/databricks/dolly-v2-12b/resolve/19308160448536e378e3db21a73a751579ee7fdd/.gitattributes",
                "expiration_timestamp": 2373918212,
            },
            "uid": "databricks/dolly-v2-12b:19308160448536e378e3db21a73a751579ee7fdd:.gitattributes",
            "file_name": "/app/model_cache/databricks_dolly_v2_12b/.gitattributes",
            "hashtype": "etag",
            "hash": "c7d9f3332a950355d5a77d85000f05e6f45435ea",
            "size": 1477,
        },
        {
            "resolution": {
                "url": "https://huggingface.co/databricks/dolly-v2-12b/resolve/19308160448536e378e3db21a73a751579ee7fdd/README.md",
                "expiration_timestamp": 2373918212,
            },
            "uid": "databricks/dolly-v2-12b:19308160448536e378e3db21a73a751579ee7fdd:README.md",
            "file_name": "/app/model_cache/databricks_dolly_v2_12b/README.md",
            "hashtype": "etag",
            "hash": "2912eb39545af0367335cff448d07214519c5eed",
            "size": 10746,
        },
        {
            "resolution": {
                "url": "https://huggingface.co/databricks/dolly-v2-12b/resolve/19308160448536e378e3db21a73a751579ee7fdd/config.json",
                "expiration_timestamp": 2373918212,
            },
            "uid": "databricks/dolly-v2-12b:19308160448536e378e3db21a73a751579ee7fdd:config.json",
            "file_name": "/app/model_cache/databricks_dolly_v2_12b/config.json",
            "hashtype": "etag",
            "hash": "888c677eda015e2375fad52d75062d14b30ebad9",
            "size": 818,
        },
        {
            "resolution": {
                "url": "https://huggingface.co/databricks/dolly-v2-12b/resolve/19308160448536e378e3db21a73a751579ee7fdd/instruct_pipeline.py",
                "expiration_timestamp": 2373918212,
            },
            "uid": "databricks/dolly-v2-12b:19308160448536e378e3db21a73a751579ee7fdd:instruct_pipeline.py",
            "file_name": "/app/model_cache/databricks_dolly_v2_12b/instruct_pipeline.py",
            "hashtype": "etag",
            "hash": "f8b291569e936cf104f44d003f95451bf5e1f965",
            "size": 9159,
        },
        {
            "resolution": {
                "url": "https://huggingface.co/databricks/dolly-v2-12b/resolve/19308160448536e378e3db21a73a751579ee7fdd/pytorch_model.bin",
                "expiration_timestamp": 2373918212,
            },
            "uid": "databricks/dolly-v2-12b:19308160448536e378e3db21a73a751579ee7fdd:pytorch_model.bin",
            "file_name": "/app/model_cache/databricks_dolly_v2_12b/pytorch_model.bin",
            "hashtype": "etag",
            "hash": "19e10711310992c310c3775964c7635f4b28dd86587403e718c6d6d524a406a5",
            "size": 23834965761,
        },
        {
            "resolution": {
                "url": "https://huggingface.co/databricks/dolly-v2-12b/resolve/19308160448536e378e3db21a73a751579ee7fdd/special_tokens_map.json",
                "expiration_timestamp": 2373918212,
            },
            "uid": "databricks/dolly-v2-12b:19308160448536e378e3db21a73a751579ee7fdd:special_tokens_map.json",
            "file_name": "/app/model_cache/databricks_dolly_v2_12b/special_tokens_map.json",
            "hashtype": "etag",
            "hash": "ecc1ee07dec13ee276fa9f1b29a1078da3280a4d",
            "size": 228,
        },
        {
            "resolution": {
                "url": "https://huggingface.co/databricks/dolly-v2-12b/resolve/19308160448536e378e3db21a73a751579ee7fdd/tokenizer.json",
                "expiration_timestamp": 2373918212,
            },
            "uid": "databricks/dolly-v2-12b:19308160448536e378e3db21a73a751579ee7fdd:tokenizer.json",
            "file_name": "/app/model_cache/databricks_dolly_v2_12b/tokenizer.json",
            "hashtype": "etag",
            "hash": "22868c8caf99a303c1a44bfea98f20f4254fc0e5",
            "size": 2114274,
        },
        {
            "resolution": {
                "url": "https://huggingface.co/databricks/dolly-v2-12b/resolve/19308160448536e378e3db21a73a751579ee7fdd/tokenizer_config.json",
                "expiration_timestamp": 2373918212,
            },
            "uid": "databricks/dolly-v2-12b:19308160448536e378e3db21a73a751579ee7fdd:tokenizer_config.json",
            "file_name": "/app/model_cache/databricks_dolly_v2_12b/tokenizer_config.json",
            "hashtype": "etag",
            "hash": "51e564ead5d28eebc74b25d86f0a694b7c7cc618",
            "size": 449,
        },
    ]
    assert len(bptr_list) == len(expected), (
        f"Expected {len(expected)} but got {len(bptr_list)}"
    )
    for expected, actual in zip(expected, bptr_list):
        assert expected["uid"] == actual.uid, (
            f"Expected uid {expected['uid']} but got {actual.uid}"
        )
        assert expected["file_name"] == actual.file_name, (
            f"Expected file_name {expected['file_name']} but got {actual.file_name}"
        )
        assert expected["hash"] == actual.hash, (
            f"Expected hash {expected['hash']} but got {actual.hash}"
        )
        assert expected["size"] == actual.size, (
            f"Expected size {expected['size']} but got {actual.size}"
        )
        assert expected["resolution"]["url"] == actual.resolution.url, (
            f"Expected resolution url {expected['resolution']['url']} but got {actual.resolution.url}"
        )
        # 100 years or more ahead
        assert (
            actual.resolution.expiration_timestamp
            >= time.time() + 20 * 365 * 24 * 60 * 60
        ), (
            f"Expected unix expiration timestamp to be at least 20 years ahead, but got {actual.resolution.expiration_timestamp}. "
        )

    # download first file and verify size
    with TemporaryDirectory() as tmp:
        # Get the first pointer (.gitattributes)
        first_pointer = bptr_list[0]
        tmp_path = Path(tmp) / "downloaded_file"

        # Download the file
        response = requests.get(first_pointer.resolution.url)
        response.raise_for_status()

        # Save the file
        tmp_path.write_bytes(response.content)

        # Verify file size matches metadata
        actual_size = tmp_path.stat().st_size
        assert actual_size == first_pointer.size, (
            f"Downloaded file size {actual_size} does not match expected size {first_pointer.size}"
        )


def test_with_main():
    # main should be resolved to 41dec486b25746052d3335decc8f5961607418a0
    cache = ModelCache(
        [
            ModelRepo(
                repo_id="intfloat/llm-retriever-base",
                revision="main",
                ignore_patterns=["*.json", "*.txt", "*.md", "*.bin", "*.model"],
                volume_folder="mistral_demo",
                use_volume=True,
            )
        ]
    )
    b10ptr = model_cache_hf_to_b10ptr(cache)
    expected = {
        "pointers": [
            {
                "resolution": {
                    "url": "https://huggingface.co/intfloat/llm-retriever-base/resolve/41dec486b25746052d3335decc8f5961607418a0/.gitattributes",
                    "expiration_timestamp": 4044816725,
                },
                "uid": "intfloat/llm-retriever-base:main:.gitattributes",
                "file_name": "/app/model_cache/mistral_demo/.gitattributes",
                "hashtype": "etag",
                "hash": "a6344aac8c09253b3b630fb776ae94478aa0275b",
                "size": 1519,
            },
            {
                "resolution": {
                    "url": "https://huggingface.co/intfloat/llm-retriever-base/resolve/41dec486b25746052d3335decc8f5961607418a0/model.safetensors",
                    "expiration_timestamp": 4044816725,
                },
                "uid": "intfloat/llm-retriever-base:main:model.safetensors",
                "file_name": "/app/model_cache/mistral_demo/model.safetensors",
                "hashtype": "etag",
                "hash": "565dd4f1cc6318ccf07af8680c27fd935b3b56ca2684d1af58abcd4e8bf6ecfa",
                "size": 437955512,
            },
        ]
    }
    assert b10ptr.model_dump() == expected


if __name__ == "__main__":
    test_dolly_12b()
    test_with_main()



import abc
import asyncio
import contextlib
import json
import logging
import threading
import time
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    ClassVar,
    Dict,
    Iterator,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
    final,
    overload,
)

import httpx
import pydantic
import tenacity

from truss.templates.shared import serialization
from truss_chains import private_types, public_types
from truss_chains import utils as chains_utils
from truss_chains.remote_chainlet import utils

if TYPE_CHECKING:
    import aiohttp


DEFAULT_DNS_CACHE_TIMEOUT_SECONDS = 300


_RetryPolicyT = TypeVar("_RetryPolicyT", tenacity.AsyncRetrying, tenacity.Retrying)
InputT = TypeVar("InputT", pydantic.BaseModel, Any)  # Any signifies "JSON".
OutputModelT = TypeVar("OutputModelT", bound=pydantic.BaseModel)


async def _safe_close(session: "aiohttp.ClientSession", timeout_sec: float) -> None:
    try:
        await asyncio.wait_for(session.close(), timeout=timeout_sec)
    except asyncio.TimeoutError:
        logging.info("Timeout while closing cycled-out aiohttp session.")
    except Exception as e:
        logging.info(f"Unexpected error while closing cycled-out aiohttp session: {e}")


class BasetenSession:
    """Provides configured HTTP clients, retries, queueing etc."""

    _client_cycle_time_sec: ClassVar[int] = 3600  # 1 hour.
    _target_url: str
    _headers: Mapping[str, str]
    _service_descriptor: public_types.DeployedServiceDescriptor
    _client_limits: httpx.Limits
    _cached_sync_client: Optional[tuple[httpx.Client, int]]
    _cached_async_client: Optional[tuple["aiohttp.ClientSession", int]]
    _sync_lock: threading.Lock
    _async_lock: asyncio.Lock
    _sync_semaphore_wrapper: utils.ThreadSemaphoreWrapper
    _async_semaphore_wrapper: utils.AsyncSemaphoreWrapper
    _close_tasks: list[asyncio.Task[None]]

    def __init__(
        self, service_descriptor: public_types.DeployedServiceDescriptor, api_key: str
    ) -> None:
        headers = {"Authorization": f"Api-Key {api_key}"}
        # If `internal_url` is present, it takes precedence.
        if service_descriptor.internal_url:
            target_msg = str(service_descriptor.internal_url)
            headers["Host"] = service_descriptor.internal_url.hostname
            target_url = service_descriptor.internal_url.gateway_run_remote_url
        elif service_descriptor.predict_url:
            target_msg = service_descriptor.predict_url
            target_url = service_descriptor.predict_url
        else:
            assert False, (
                "Per validation of `DeployedServiceDescriptor` either `predict_url` "
                f"or `internal_url` must be present. Got {service_descriptor}"
            )

        logging.info(
            f"Creating BasetenSession (HTTP) for `{service_descriptor.name}`.\n"
            f"\tTarget: `{target_msg}`\n\t`{service_descriptor.options}`."
        )
        self._target_url = target_url
        self._headers = headers
        self._service_descriptor = service_descriptor
        self._client_limits = httpx.Limits(
            max_connections=service_descriptor.options.concurrency_limit,
            max_keepalive_connections=50,
        )
        self._cached_sync_client = None
        self._cached_async_client = None
        self._sync_lock = threading.Lock()
        self._async_lock = asyncio.Lock()
        self._sync_semaphore_wrapper = utils.ThreadSemaphoreWrapper(
            service_descriptor.options.concurrency_limit, service_descriptor.name
        )
        self._async_semaphore_wrapper = utils.AsyncSemaphoreWrapper(
            service_descriptor.options.concurrency_limit, service_descriptor.name
        )
        self._close_tasks = []

    @property
    def name(self) -> str:
        return self._service_descriptor.name

    async def shut_down(self) -> None:
        # TODO: integrate this with the uvicorn server.
        await asyncio.gather(*self._close_tasks)

    def _client_cycle_needed(self, cached_client: Optional[tuple[Any, int]]) -> bool:
        return (
            not cached_client
            or (int(time.time()) - cached_client[1]) > self._client_cycle_time_sec
        )

    def _log_retry(self, retry_state: tenacity.RetryCallState) -> None:
        logging.info(f"Retrying `{self.name}`, attempt {retry_state.attempt_number}")

    def _make_retry_policy(self, retrying: Type[_RetryPolicyT]) -> _RetryPolicyT:
        return retrying(
            stop=tenacity.stop_after_attempt(self._service_descriptor.options.retries),
            retry=tenacity.retry_if_exception_type(Exception),
            reraise=True,
            before_sleep=self._log_retry,
        )

    @contextlib.contextmanager
    def _client_sync(self) -> Iterator[httpx.Client]:
        # Check `_client_cycle_needed` before and after locking to avoid
        # needing a lock each time the client is accessed.
        if self._client_cycle_needed(self._cached_sync_client):
            with self._sync_lock:
                if self._client_cycle_needed(self._cached_sync_client):
                    self._cached_sync_client = (
                        httpx.Client(
                            headers=self._headers,
                            timeout=self._service_descriptor.options.timeout_sec,
                            limits=self._client_limits,
                        ),
                        int(time.time()),
                    )
        assert self._cached_sync_client is not None
        client = self._cached_sync_client[0]

        with self._sync_semaphore_wrapper():
            yield client

    @contextlib.asynccontextmanager
    async def _client_async(self) -> AsyncIterator["aiohttp.ClientSession"]:
        try:
            import aiohttp
        except ImportError:
            raise chains_utils.make_optional_import_error("aiohttp")

        # Check `_client_cycle_needed` before and after locking to avoid
        # needing a lock each time the client is accessed.
        if self._client_cycle_needed(self._cached_async_client):
            async with self._async_lock:
                if self._client_cycle_needed(self._cached_async_client):
                    if self._cached_async_client is not None:
                        # Close with same timeout as connections, but add some buffer.
                        self._close_tasks.append(
                            asyncio.create_task(
                                _safe_close(
                                    self._cached_async_client[0],
                                    self._service_descriptor.options.timeout_sec * 1.1,
                                )
                            )
                        )
                    limit = self._client_limits.max_connections
                    assert limit is not None
                    connector = aiohttp.TCPConnector(
                        limit=limit, ttl_dns_cache=DEFAULT_DNS_CACHE_TIMEOUT_SECONDS
                    )
                    self._cached_async_client = (
                        aiohttp.ClientSession(
                            headers=self._headers,
                            connector=connector,
                            timeout=aiohttp.ClientTimeout(
                                total=self._service_descriptor.options.timeout_sec
                            ),
                        ),
                        int(time.time()),
                    )
        assert self._cached_async_client is not None
        client = self._cached_async_client[0]

        async with self._async_semaphore_wrapper():
            yield client


class StubBase(BasetenSession, abc.ABC):
    """Base class for stubs that invoke remote chainlets.

    Extends ``BasetenSession`` with methods for data serialization, de-serialization
    and invoking other endpoints.

    It is used internally for RPCs to dependency chainlets, but it can also be used
    in user-code for wrapping a deployed truss model into the Chains framework. It
    flexibly supports JSON and pydantic inputs and output. Example usage::

        import pydantic
        import truss_chains as chains


        class WhisperOutput(pydantic.BaseModel):
            ...


        class DeployedWhisper(chains.StubBase):
            # Input JSON, output JSON.
            async def run_remote(self, audio_b64: str) -> Any:
                return await self.predict_async(
                    inputs={"audio": audio_b64})
                # resp == {"text": ..., "language": ...}

            # OR Input JSON, output pydantic model.
            async def run_remote(self, audio_b64: str) -> WhisperOutput:
                return await self.predict_async(
                    inputs={"audio": audio_b64}, output_model=WhisperOutput)

            # OR Input and output are pydantic models.
            async def run_remote(self, data: WhisperInput) -> WhisperOutput:
                return await self.predict_async(data, output_model=WhisperOutput)


        class MyChainlet(chains.ChainletBase):

            def __init__(self, ..., context=chains.depends_context()):
                ...
                self._whisper = DeployedWhisper.from_url(
                    WHISPER_URL,
                    context,
                    options=chains.RPCOptions(retries=3),
                )

            async def run_remote(self, ...):
               await self._whisper.run_remote(...)
    """

    @final
    def __init__(
        self, service_descriptor: public_types.DeployedServiceDescriptor, api_key: str
    ) -> None:
        """
        Args:
            service_descriptor: Contains the URL and other configuration.
            api_key: A baseten API key to authorize requests.
        """
        super().__init__(service_descriptor, api_key)

    @classmethod
    def from_url(
        cls,
        predict_url: str,
        context_or_api_key: Union[public_types.DeploymentContext, str],
        options: Optional[public_types.RPCOptions] = None,
    ):
        """Factory method, convenient to be used in chainlet's ``__init__``-method.

        Args:
            predict_url: URL to predict endpoint of another chain / truss model.
            context_or_api_key: Deployment context object, obtained in the
               chainlet's ``__init__`` or Baseten API key.
            options: RPC options, e.g. retries.
        """
        options = options or public_types.RPCOptions()
        if isinstance(context_or_api_key, str):
            api_key = context_or_api_key
        else:
            api_key = context_or_api_key.get_baseten_api_key()
        return cls(
            service_descriptor=public_types.DeployedServiceDescriptor(
                name=cls.__name__,
                display_name=cls.__name__,
                predict_url=predict_url,
                options=options,
            ),
            api_key=api_key,
        )

    def _make_request_params(
        self, inputs: InputT, for_httpx: bool = False
    ) -> Mapping[str, Any]:
        kwargs: Dict[str, Any] = {}
        headers = {}
        if trace_parent := utils.get_trace_parent():
            headers[private_types.OTEL_TRACE_PARENT_HEADER_KEY] = trace_parent

        if isinstance(inputs, pydantic.BaseModel):
            if self._service_descriptor.options.use_binary:
                data_dict = inputs.model_dump(mode="python")
                data_key = "content" if for_httpx else "data"
                kwargs[data_key] = serialization.truss_msgpack_serialize(data_dict)
                headers["Content-Type"] = "application/octet-stream"
            else:
                data_key = "content" if for_httpx else "data"
                kwargs[data_key] = inputs.model_dump_json()
                headers["Content-Type"] = "application/json"
        else:  # inputs is JSON dict.
            if self._service_descriptor.options.use_binary:
                data_key = "content" if for_httpx else "data"
                kwargs[data_key] = serialization.truss_msgpack_serialize(inputs)
                headers["Content-Type"] = "application/octet-stream"
            else:
                kwargs["json"] = inputs
                headers["Content-Type"] = "application/json"

        kwargs["headers"] = headers
        return kwargs

    def _response_to_pydantic(
        self, response: bytes, output_model: Type[OutputModelT]
    ) -> OutputModelT:
        if self._service_descriptor.options.use_binary:
            data_dict = serialization.truss_msgpack_deserialize(response)
            return output_model.model_validate(data_dict)
        return output_model.model_validate_json(response)

    def _response_to_json(self, response: bytes) -> Any:
        if self._service_descriptor.options.use_binary:
            return serialization.truss_msgpack_deserialize(response)
        return json.loads(response)

    @overload
    def predict_sync(
        self, inputs: InputT, output_model: Type[OutputModelT]
    ) -> OutputModelT:
        """Returns a validated pydantic model. Inputs can be pydantic or JSON dict."""

    @overload
    def predict_sync(self, inputs: InputT, output_model: None = None) -> Any:
        """Returns a raw JSON dict. Inputs can be pydantic or JSON dict."""

    def predict_sync(
        self, inputs: InputT, output_model: Optional[Type[OutputModelT]] = None
    ) -> Union[OutputModelT, Any]:
        retry = self._make_retry_policy(tenacity.Retrying)
        params = self._make_request_params(inputs, for_httpx=True)

        def _rpc() -> bytes:
            client: httpx.Client
            with self._client_sync() as client:
                response = client.post(self._target_url, **params)
            utils.response_raise_errors(response, self.name)
            return response.content

        try:
            response_bytes = retry(_rpc)
        except httpx.ReadTimeout:
            msg = (
                f"Timeout calling remote Chainlet `{self.name}` "
                f"({self._service_descriptor.options.timeout_sec} seconds limit)."
            )
            logging.warning(msg)
            raise TimeoutError(msg) from None  # Prune error stack trace (TMI).

        if output_model:
            return self._response_to_pydantic(response_bytes, output_model)
        return self._response_to_json(response_bytes)

    @overload
    async def predict_async(
        self, inputs: InputT, output_model: Type[OutputModelT]
    ) -> OutputModelT:
        """Returns a validated pydantic model. Inputs can be pydantic or JSON dict."""

    @overload
    async def predict_async(self, inputs: InputT, output_model: None = None) -> Any:
        """Returns a validated pydantic model. Inputs can be pydantic or JSON dict."""

    async def predict_async(
        self, inputs: InputT, output_model: Optional[Type[OutputModelT]] = None
    ) -> Union[OutputModelT, Any]:
        retry = self._make_retry_policy(tenacity.AsyncRetrying)
        params = self._make_request_params(inputs)

        async def _rpc() -> bytes:
            client: "aiohttp.ClientSession"
            async with self._client_async() as client:
                async with client.post(self._target_url, **params) as response:
                    await utils.async_response_raise_errors(response, self.name)
                    return await response.read()

        try:
            response_bytes: bytes = await retry(_rpc)
        except asyncio.TimeoutError:
            msg = (
                f"Timeout calling remote Chainlet `{self.name}` "
                f"({self._service_descriptor.options.timeout_sec} seconds limit)."
            )
            logging.warning(msg)
            raise TimeoutError(msg) from None  # Prune error stack trace (TMI).

        if output_model:
            return self._response_to_pydantic(response_bytes, output_model)
        return self._response_to_json(response_bytes)

    async def predict_async_stream(self, inputs: InputT) -> AsyncIterator[bytes]:
        retry = self._make_retry_policy(tenacity.AsyncRetrying)
        params = self._make_request_params(inputs)

        async def _rpc() -> AsyncIterator[bytes]:
            client: "aiohttp.ClientSession"
            async with self._client_async() as client:
                response = await client.post(self._target_url, **params)
                await utils.async_response_raise_errors(response, self.name)
                return response.content.iter_any()

        try:
            return await retry(_rpc)
        except asyncio.TimeoutError:
            msg = (
                f"Timeout calling remote Chainlet `{self.name}` "
                f"({self._service_descriptor.options.timeout_sec} seconds limit)."
            )
            logging.warning(msg)
            raise TimeoutError(msg) from None  # Prune error stack trace (TMI).


StubT = TypeVar("StubT", bound=StubBase)


def factory(stub_cls: Type[StubT], context: public_types.DeploymentContext) -> StubT:
    # Assumes the stub_cls-name and the name of the service in ``context` match.
    return stub_cls(
        service_descriptor=context.get_service_descriptor(stub_cls.__name__),
        api_key=context.get_baseten_api_key(),
    )



import os
import time
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
import requests
from python_on_whales.exceptions import DockerException
from tenacity import RetryError

from truss.base.custom_types import Example
from truss.base.errors import ContainerIsDownError, ContainerNotFoundError
from truss.base.truss_config import map_local_to_supported_python_version
from truss.local.local_config_handler import LocalConfigHandler
from truss.templates.control.control.helpers.custom_types import (
    Action,
    ModelCodePatch,
    Patch,
    PatchType,
)
from truss.tests.test_testing_utilities_for_other_tests import (
    ensure_kill_all,
    kill_all_with_retries,
)
from truss.truss_handle.patch.custom_types import PatchRequest
from truss.truss_handle.truss_handle import DockerURLs, TrussHandle, wait_for_truss
from truss.util.docker import Docker, DockerStates


def test_spec(custom_model_truss_dir_with_pre_and_post):
    dir_path = custom_model_truss_dir_with_pre_and_post
    th = TrussHandle(dir_path)
    spec = th.spec
    assert spec.truss_dir == dir_path


def test_description(custom_model_truss_dir_with_pre_and_post_description):
    dir_path = custom_model_truss_dir_with_pre_and_post_description
    th = TrussHandle(dir_path)
    spec = th.spec
    assert spec.description == "This model adds 3 to all inputs"


def test_predict(custom_model_truss_dir_with_pre_and_post):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    resp = th.predict([1, 2, 3, 4])
    assert resp == {"predictions": [4, 5, 6, 7]}


def test_predict_with_external_packages(custom_model_with_external_package):
    th = TrussHandle(custom_model_with_external_package)
    resp = th.predict([1, 2, 3, 4])
    assert resp == [1, 1, 1, 1]


def test_server_predict(custom_model_truss_dir_with_pre_and_post):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    resp = th.server_predict([1, 2, 3, 4])
    assert resp == {"predictions": [4, 5, 6, 7]}


def test_readme_generation_int_example(
    test_data_path, custom_model_truss_dir_with_pre_and_post
):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    readme_contents = th.generate_readme()
    readme_contents = readme_contents.replace("\n", "")
    correct_readme_contents = _read_readme(test_data_path / "readme_int_example.md")
    assert readme_contents == correct_readme_contents


def test_readme_generation_no_example(
    test_data_path, custom_model_truss_dir_with_pre_and_post_no_example
):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post_no_example)
    if os.path.exists(th._spec.examples_path):
        # Remove the examples file
        os.remove(th._spec.examples_path)
    readme_contents = th.generate_readme()
    readme_contents = readme_contents.replace("\n", "")
    correct_readme_contents = _read_readme(test_data_path / "readme_no_example.md")
    assert readme_contents == correct_readme_contents


def test_readme_generation_str_example(
    test_data_path, custom_model_truss_dir_with_pre_and_post_str_example
):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post_str_example)
    readme_contents = th.generate_readme()
    readme_contents = readme_contents.replace("\n", "")
    correct_readme_contents = _read_readme(test_data_path / "readme_str_example.md")
    assert readme_contents == correct_readme_contents


@pytest.mark.integration
def test_build_docker_image(custom_model_truss_dir_with_pre_and_post):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    tag = "test-build-image-tag:0.0.1"
    image = th.build_serving_docker_image(tag=tag)
    assert image.repo_tags[0] == tag


@pytest.mark.integration
@pytest.mark.parametrize(
    "base_image, path, expected_fail",
    [
        ("baseten/truss-server-base:3.9-v0.4.8rc4", "/usr/local/bin/python3", False),
        ("python:3.8", "/usr/local/bin/python3", False),
        ("python:3.10", "/usr/local/bin/python3", False),
        ("python:3.11", "/usr/local/bin/python3", False),
        ("python:3.13", "/usr/local/bin/python3", False),
        ("python:alpine", "/usr/local/bin/python3", True),
        ("python:2.7-slim", "/usr/local/bin/python", True),
        ("python:3.7-slim", "/usr/local/bin/python3", True),
    ],
)
def test_build_serving_docker_image_from_user_base_image_live_reload(
    custom_model_truss_dir, base_image, path, expected_fail
):
    th = TrussHandle(custom_model_truss_dir)
    th.set_base_image(base_image, path)
    th.live_reload()
    try:
        th.build_serving_docker_image(cache=False)
    except DockerException as exc:
        assert expected_fail is True
        assert "It returned with code 1" in str(exc)


@pytest.mark.integration
def test_docker_predict_custom_base_image(custom_model_truss_dir_with_pre_and_post):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    th.set_base_image(
        "wallies/python-cuda:3.10-cuda11.7-runtime", "/usr/local/bin/python"
    )
    with ensure_kill_all():
        result = th.docker_predict([1, 2], local_port=None)
        assert result == {"predictions": [4, 5]}


@pytest.mark.integration
def test_build_docker_image_gpu(custom_model_truss_dir_for_gpu, tmp_path):
    th = TrussHandle(custom_model_truss_dir_for_gpu)
    tag = "test-build-image-gpu-tag:0.0.1"
    build_dir = tmp_path / "scaffold_build_dir"
    image = th.build_serving_docker_image(tag=tag, build_dir=build_dir)
    assert image.repo_tags[0] == tag


@pytest.mark.integration
def test_build_docker_image_control_gpu(custom_model_truss_dir_for_gpu, tmp_path):
    th = TrussHandle(custom_model_truss_dir_for_gpu)
    th.live_reload(True)
    tag = "test-build-image-control-gpu-tag:0.0.1"
    build_dir = tmp_path / "scaffold_build_dir"
    image = th.build_serving_docker_image(tag=tag, build_dir=build_dir)
    assert image.repo_tags[0] == tag


@pytest.mark.integration
def test_docker_run(custom_model_truss_dir_with_pre_and_post):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    tag = "test-docker-run-tag:0.0.1"
    container = th.docker_run(tag=tag, local_port=None)
    try:
        assert _container_exists(container)
    finally:
        Docker.client().kill(container)


@pytest.mark.skip(reason="Needs gpu")
@pytest.mark.integration
def test_docker_run_gpu(custom_model_truss_dir_for_gpu):
    th = TrussHandle(custom_model_truss_dir_for_gpu)
    tag = "test-docker-run-gpu-tag:0.0.1"
    container = th.docker_run(tag=tag, local_port=None)
    try:
        assert _container_exists(container)
    finally:
        Docker.client().kill(container)


@pytest.mark.integration
def test_docker_run_without_tag(custom_model_truss_dir_with_pre_and_post):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    container = th.docker_run(local_port=None)
    try:
        assert _container_exists(container)
    finally:
        Docker.client().kill(container)


@pytest.mark.integration
def get_docker_containers_from_labels(custom_model_truss_dir_with_pre_and_post):
    with ensure_kill_all():
        t1 = TrussHandle(custom_model_truss_dir_with_pre_and_post)
        assert len(t1.get_serving_docker_containers_from_labels()) == 0
        t1.docker_run(local_port=None)
        assert len(t1.get_serving_docker_containers_from_labels()) == 1
        t1.docker_run(local_port=None)
        assert len(t1.get_serving_docker_containers_from_labels()) == 2
        t1.kill_container()
        assert len(t1.get_serving_docker_containers_from_labels()) == 0


@pytest.mark.integration
def test_predict_use_docker(custom_model_truss_dir_with_pre_and_post):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    tag = "test-docker-predict-tag:0.0.1"
    with ensure_kill_all():
        result = th.predict([1, 2], tag=tag, use_docker=True)
        assert result == {"predictions": [4, 5]}


@pytest.mark.integration
def test_docker_predict(custom_model_truss_dir_with_pre_and_post):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    tag = "test-docker-predict-tag:0.0.1"
    with ensure_kill_all():
        result = th.docker_predict([1, 2], tag=tag, local_port=None)
        assert result == {"predictions": [4, 5]}


@pytest.mark.integration
def test_docker_predict_model_with_external_packages(
    custom_model_with_external_package,
):
    th = TrussHandle(custom_model_with_external_package)
    tag = "test-docker-predict-ext-pkg-tag:0.0.1"
    with ensure_kill_all():
        result = th.docker_predict([1, 2], tag=tag, local_port=None)
        assert result == [1, 1]


@pytest.mark.integration
def test_docker_predict_with_bundled_packages(
    custom_model_truss_dir_with_bundled_packages,
):
    th = TrussHandle(custom_model_truss_dir_with_bundled_packages)
    tag = "test-docker-predict-bundled-packages-tag:0.0.1"
    with ensure_kill_all():
        result = th.docker_predict([1, 2], tag=tag, local_port=None)
        assert result == {"predictions": [1]}


@pytest.mark.integration
def test_docker_multiple_predict(custom_model_truss_dir_with_pre_and_post):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    tag = "test-docker-predict-tag:0.0.1"
    with ensure_kill_all():
        r1 = th.docker_predict([1, 2], tag=tag, local_port=None)
        r2 = th.docker_predict([3, 4], tag=tag, local_port=None)
        assert r1 == {"predictions": [4, 5]}
        assert r2 == {"predictions": [6, 7]}
        assert len(th.get_serving_docker_containers_from_labels()) == 1


@pytest.mark.integration
def test_kill_all(custom_model_truss_dir, custom_model_truss_dir_with_pre_and_post):
    t1 = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    t2 = TrussHandle(custom_model_truss_dir)
    with ensure_kill_all():
        t1.docker_run(local_port=None)
        assert len(t1.get_serving_docker_containers_from_labels()) == 1
        t2.docker_run(local_port=None)
        assert len(t2.get_serving_docker_containers_from_labels()) == 1
        kill_all_with_retries()
        assert len(t1.get_serving_docker_containers_from_labels()) == 0
        assert len(t2.get_serving_docker_containers_from_labels()) == 0


@pytest.mark.skip(reason="Needs gpu")
@pytest.mark.integration
def test_docker_predict_gpu(custom_model_truss_dir_for_gpu):
    th = TrussHandle(custom_model_truss_dir_for_gpu)
    tag = "test-docker-predict-gpu-tag:0.0.1"
    with ensure_kill_all():
        result = th.docker_predict([1], tag=tag, local_port=None)
        assert result["predictions"][0]["cuda_version"].startswith("11")


@pytest.mark.integration
def test_docker_predict_secrets(custom_model_truss_dir_for_secrets):
    th = TrussHandle(custom_model_truss_dir_for_secrets)
    tag = "test-docker-predict-secrets-tag:0.0.1"
    LocalConfigHandler.set_secret("secret_name", "secret_value")
    with ensure_kill_all():
        try:
            result = th.docker_predict(
                {"instances": ["secret_name"]}, tag=tag, local_port=None
            )
            assert result["predictions"][0] == "secret_value"
        finally:
            LocalConfigHandler.remove_secret("secret_name")


@pytest.mark.integration
def test_docker_no_preprocess_custom_model(no_preprocess_custom_model):
    th = TrussHandle(no_preprocess_custom_model)
    tag = "test-docker-no-preprocess-tag:0.0.1"
    with ensure_kill_all():
        result = th.docker_predict([1], tag=tag, local_port=None)
        assert result["predictions"][0] == 2


@pytest.mark.integration
def test_docker_long_load(long_load_model):
    th = TrussHandle(long_load_model)
    tag = "test-docker-long-load-tag:0.0.1"
    with ensure_kill_all():
        result = th.docker_predict([1], tag=tag, local_port=None)
        assert result["predictions"][0] == 1


@pytest.mark.integration
def test_local_no_preprocess_custom_model(no_preprocess_custom_model):
    th = TrussHandle(no_preprocess_custom_model)
    result = th.server_predict([1])
    assert result["predictions"][0] == 2


@pytest.mark.integration
def test_docker_no_postprocess_custom_model(no_postprocess_custom_model):
    th = TrussHandle(no_postprocess_custom_model)
    tag = "test-docker-no-postprocess-tag:0.0.1"
    with ensure_kill_all():
        result = th.docker_predict([1], tag=tag, local_port=None)
        assert result["predictions"][0] == 2


@pytest.mark.integration
def test_local_no_postprocess_custom_model(no_postprocess_custom_model):
    th = TrussHandle(no_postprocess_custom_model)
    result = th.server_predict([1])
    assert result["predictions"][0] == 2


@pytest.mark.integration
def test_docker_no_load_custom_model(no_load_custom_model):
    th = TrussHandle(no_load_custom_model)
    tag = "test-docker-no-load-tag:0.0.1"
    with ensure_kill_all():
        result = th.docker_predict([1], tag=tag, local_port=None)
        assert result["predictions"][0] == 1


@pytest.mark.integration
def test_local_no_load_custom_model(no_load_custom_model):
    th = TrussHandle(no_load_custom_model)
    result = th.server_predict([1])
    assert result["predictions"][0] == 1


@pytest.mark.integration
def test_docker_no_params_init_custom_model(no_params_init_custom_model):
    th = TrussHandle(no_params_init_custom_model)
    tag = "test-docker-no-params-init-tag:0.0.1"
    with ensure_kill_all():
        result = th.docker_predict([1], tag=tag, local_port=None)
        assert result["predictions"][0] == 1


@pytest.mark.integration
def test_local_no_params_init_custom_model(no_params_init_custom_model):
    th = TrussHandle(no_params_init_custom_model)
    result = th.server_predict([1])
    assert result["predictions"][0] == 1


@pytest.mark.integration
def test_custom_python_requirement(custom_model_truss_dir_with_pre_and_post):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    th.add_python_requirement("theano")
    th.add_python_requirement("scipy")
    tag = "test-custom-python-req-tag:0.0.1"
    container = th.docker_run(tag=tag, local_port=None)
    try:
        verify_python_requirement_installed_on_container(container, "theano")
        verify_python_requirement_installed_on_container(container, "scipy")
    finally:
        Docker.client().kill(container)


@pytest.mark.integration
def test_custom_system_package(custom_model_truss_dir_with_pre_and_post):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    th.add_system_package("jq")
    th.add_system_package("fzf")
    tag = "test-custom-system-package-tag:0.0.1"
    container = th.docker_run(tag=tag, local_port=None)
    try:
        verify_system_package_installed_on_container(container, "jq")
        verify_system_package_installed_on_container(container, "fzf")
    finally:
        Docker.client().kill(container)


@pytest.mark.parametrize(
    "python_version, expected_python_version",
    [("3.8", "py38"), ("py38", "py38"), ("3.9", "py39"), ("py39", "py39")],
)
def test_update_python_version(
    python_version, expected_python_version, custom_model_truss_dir_with_pre_and_post
):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    th.update_python_version(python_version)
    assert th.spec.python_version == expected_python_version


def test_update_requirements(custom_model_truss_dir_with_pre_and_post):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    requirements = ["tensorflow==2.3.1", "uvicorn==0.12.2"]
    th.update_requirements(requirements)
    sc_requirements = th.spec.requirements
    assert sc_requirements == requirements


def test_update_requirements_from_file(
    custom_model_truss_dir_with_pre_and_post, tmp_path
):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    file_requirements = [
        "tensorflow==2.3.1",
        "# this is comment. Please don't add.",
        "    # this is comment with a big space. Please don't add.",
        "uvicorn==0.12.2",
    ]
    allowed_requirements = ["tensorflow==2.3.1", "uvicorn==0.12.2"]
    req_file_path = tmp_path / "requirements.txt"
    with req_file_path.open("w") as req_file:
        for req in file_requirements:
            req_file.write(f"{req}\n")
    th.update_requirements_from_file(str(req_file_path))
    sc_requirements = th.spec.requirements
    assert sc_requirements == allowed_requirements


@pytest.mark.integration
def test_add_environment_variable(custom_model_truss_dir_with_pre_and_post):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    th.add_environment_variable("test_env", "test_value")
    tag = "test-add-env-var-tag:0.0.1"
    container = th.docker_run(tag=tag, local_port=None)
    try:
        verify_environment_variable_on_container(container, "test_env", "test_value")
    finally:
        Docker.client().kill(container)


@pytest.mark.integration
def test_build_commands(test_data_path):
    truss_dir = test_data_path / "test_build_commands"
    tr = TrussHandle(truss_dir)
    with ensure_kill_all():
        r1 = tr.docker_predict([1, 2], local_port=None)
        assert r1 == {"predictions": [1, 2]}


@pytest.mark.integration
def test_build_commands_failure(test_data_path):
    truss_dir = test_data_path / "test_build_commands_failure"
    tr = TrussHandle(truss_dir)
    try:
        tr.docker_run(local_port=None, detach=True, wait_for_server_ready=True)
    except DockerException as exc:
        assert "It returned with code 1" in str(exc)


def test_add_data_file(custom_model_truss_dir_with_pre_and_post, tmp_path):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    data_filepath = tmp_path / "test_data.txt"
    with data_filepath.open("w") as data_file:
        data_file.write("test")

    th.add_data(str(data_filepath))

    scaf_data_filepath = th.spec.data_dir / "test_data.txt"
    assert scaf_data_filepath.exists()
    with scaf_data_filepath.open() as data_file:
        assert data_file.read() == "test"


def test_add_data_fileglob(custom_model_truss_dir_with_pre_and_post, tmp_path):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    file1_path = tmp_path / "test_data1.txt"
    with file1_path.open("w") as data_file:
        data_file.write("test1")

    file2_path = tmp_path / "test_data2.txt"
    with file2_path.open("w") as data_file:
        data_file.write("test2")

    file2_path = tmp_path / "test_data3.json"
    with file2_path.open("w") as data_file:
        data_file.write("{}")

    th.add_data(f"{str(tmp_path)}/*.txt")

    assert (th.spec.data_dir / "test_data1.txt").exists()
    assert (th.spec.data_dir / "test_data2.txt").exists()
    assert not (th.spec.data_dir / "test_data2.json").exists()


def test_add_data_dir(custom_model_truss_dir_with_pre_and_post, tmp_path):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    sub_dir = tmp_path / "sub"
    sub_sub_dir = sub_dir / "sub"
    sub_sub_dir.mkdir(parents=True)

    file_path = sub_sub_dir / "test_file.txt"
    with file_path.open("w") as data_file:
        data_file.write("test")

    th.add_data(str(sub_dir))

    scaf_file_path = th.spec.data_dir / "sub" / "sub" / "test_file.txt"
    assert scaf_file_path.exists()
    with scaf_file_path.open() as data_file:
        assert data_file.read() == "test"


def test_examples(custom_model_truss_dir_with_pre_and_post):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    examples = th.examples()
    assert "example1" in [example.name for example in examples]


def test_add_example_new(custom_model_truss_dir_with_pre_and_post):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    orig_examples = th.examples()
    th.add_example("example2", [[1]])
    assert th.examples() == [*orig_examples, Example("example2", [[1]])]


def test_add_example_update(custom_model_truss_dir_with_pre_and_post):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)
    th.add_example("example1", [[1]])
    assert th.examples() == [Example("example1", [[1]])]


def test_model_without_pre_post(custom_model_truss_dir):
    th = TrussHandle(custom_model_truss_dir)
    resp = th.server_predict([1, 2, 3, 4])
    assert resp == [1, 1, 1, 1]


@pytest.mark.integration
def test_docker_predict_model_without_pre_post(custom_model_truss_dir):
    th = TrussHandle(custom_model_truss_dir)
    with ensure_kill_all():
        resp = th.docker_predict([1, 2, 3, 4], local_port=None)
        assert resp == [1, 1, 1, 1]


@pytest.mark.integration
def test_control_truss_apply_patch(custom_model_control):
    th = TrussHandle(custom_model_control)
    tag = "test-docker-custom-model-control-tag:0.0.1"
    with ensure_kill_all():
        result = th.docker_predict([1], tag=tag, local_port=None)
        assert result[0] == 1

        running_hash = th.truss_hash_on_serving_container()
        new_model_code = """
class Model:
    def predict(self, model_input):
        return [2 for i in model_input]
"""
        patch_request = PatchRequest(
            hash="dummy",
            prev_hash=running_hash,
            patches=[
                Patch(
                    type=PatchType.MODEL_CODE,
                    body=ModelCodePatch(
                        action=Action.UPDATE, path="model.py", content=new_model_code
                    ),
                )
            ],
        )

        th.patch_container(patch_request)
        result = th.docker_predict([1], tag=tag, local_port=None)
        assert result[0] == 2


@pytest.mark.integration
def test_regular_truss_local_update_flow(custom_model_truss_dir):
    th = TrussHandle(custom_model_truss_dir)
    tag = "test-docker-custom-model-tag:0.0.1"
    with ensure_kill_all():
        result = th.docker_predict([1], tag=tag, local_port=None)
        assert result[0] == 1
        orig_num_truss_images = len(th.get_all_docker_images())

        # No new docker images on second predict
        result = th.docker_predict([1], tag=tag, local_port=None)
        assert orig_num_truss_images == len(th.get_all_docker_images())

        with (custom_model_truss_dir / "model" / "model.py").open(
            "w"
        ) as model_code_file:
            model_code_file.write(
                """
class Model:
    def predict(self, model_input):
        return [2 for i in model_input]
"""
            )
        result = th.docker_predict([1], tag=tag, local_port=None)
        assert result[0] == 2
        # A new image should have been created
        assert len(th.get_all_docker_images()) == orig_num_truss_images + 1


@patch("truss.truss_handle.truss_handle.directory_content_hash")
def test_truss_hash_caching_based_on_max_mod_time(
    directory_content_patcher, custom_model_truss_dir
):
    directory_content_patcher.return_value = "mock_hash"
    th = TrussHandle(custom_model_truss_dir)
    labels = th._get_serving_labels()
    labels2 = th._get_serving_labels()
    assert labels == labels2
    directory_content_patcher.assert_called_once()

    time.sleep(0.1)  # Make sure different mod time
    (custom_model_truss_dir / "model" / "model.py").touch()
    labels3 = th._get_serving_labels()
    assert labels3 != labels
    directory_content_patcher.call_count == 2


@patch("truss.truss_handle.truss_handle.get_docker_urls")
@patch("truss.truss_handle.truss_handle.get_container_state")
def test_container_oom_caught_during_waiting(
    container_state_mock, get_docker_urls_mock
):
    container_state_mock.return_value = DockerStates.OOMKILLED
    get_docker_urls_mock.return_value = DockerURLs("http://localhost:8080")
    with pytest.raises(ContainerIsDownError):
        wait_for_truss(container=MagicMock())


@patch("truss.truss_handle.truss_handle.get_docker_urls")
@patch("truss.truss_handle.truss_handle.get_container_state")
@pytest.mark.integration
def test_container_stuck_in_created(container_state_mock, get_docker_urls_mock):
    container_state_mock.return_value = DockerStates.CREATED
    get_docker_urls_mock.return_value = DockerURLs("http://localhost:8080")
    with pytest.raises(ContainerIsDownError):
        wait_for_truss(container=MagicMock())


@pytest.mark.integration
def test_control_truss_local_update_that_crashes_inference_server(custom_model_control):
    th = TrussHandle(custom_model_control)
    tag = "test-docker-custom-model-control-tag:0.0.1"
    with ensure_kill_all():
        result = th.docker_predict([1], tag=tag, local_port=None)
        assert result[0] == 1

        bad_model_code = """
class Model:
    def malformed
"""
        model_code_file_path = custom_model_control / "model" / "model.py"
        with model_code_file_path.open("w") as model_code_file:
            model_code_file.write(bad_model_code)
        with pytest.raises(RetryError) as exc_info:
            th.docker_predict([1], tag=tag, local_port=None)
        resp = exc_info.value.last_attempt.result()
        assert resp.status_code == 503
        assert (
            "Model load failed" in resp.text
            or "It appears your model has stopped running" in resp.text
        )

        # Should be able to fix code after
        good_model_code = """
class Model:
    def predict(self, model_input):
        return [2 for i in model_input]
"""
        with model_code_file_path.open("w") as model_code_file:
            model_code_file.write(good_model_code)
        result = th.docker_predict([1], tag=tag, local_port=None)
        assert result[0] == 2


@pytest.mark.integration
@pytest.mark.parametrize(
    "patch_path, expected_call_count",
    [("hash_is_current", 1), ("hash_is_current_but_only_every_third_call_succeeds", 3)],
)
def test_patch_ping_flow(
    patch_path, expected_call_count, custom_model_control, patch_ping_test_server
):
    port = patch_ping_test_server
    patch_ping_url = f"http://host.docker.internal:{port}/{patch_path}"
    th = TrussHandle(custom_model_control)
    tag = "test-docker-custom-model-control-tag:0.0.1"
    with ensure_kill_all():
        result = th.docker_predict(
            [1], tag=tag, patch_ping_url=patch_ping_url, local_port=None
        )
        assert result == [1]

        # Make sure the patch ping url was actually hit
        stats = requests.get(f"http://127.0.0.1:{port}/stats").json()
        assert stats[f"{patch_path}_called_count"] == expected_call_count


def test_handle_if_container_dne(custom_model_truss_dir):
    def return_container_dne(self):
        return "DNE"

    with (
        patch.object(TrussHandle, "_try_patch", new=return_container_dne),
        pytest.raises(ContainerNotFoundError),
    ):
        truss_handle = TrussHandle(truss_dir=custom_model_truss_dir)
        truss_handle.docker_run(local_port=None)
    kill_all_with_retries()


def test_docker_predict_container_does_not_exist(custom_model_truss_dir):
    def return_container_dne(self):
        return "DNE"

    with (
        patch.object(TrussHandle, "_try_patch", new=return_container_dne),
        pytest.raises(ContainerNotFoundError),
    ):
        truss_handle = TrussHandle(truss_dir=custom_model_truss_dir)
        truss_handle.docker_predict([1], local_port=None)
    kill_all_with_retries()


@pytest.mark.integration
def test_external_data(custom_model_external_data_access_tuple_fixture):
    truss_dir, expected_content = custom_model_external_data_access_tuple_fixture
    th = TrussHandle(truss_dir)
    tag = "test-external-data-access-tag:0.0.1"
    with ensure_kill_all():
        result = th.docker_predict([], tag=tag, network="host", local_port=None)
        assert result == expected_content


@pytest.mark.integration
def test_external_data_gpu(custom_model_external_data_access_tuple_fixture_gpu):
    truss_dir, expected_content = custom_model_external_data_access_tuple_fixture_gpu
    th = TrussHandle(truss_dir)
    tag = "test-external-data-access-tag:0.0.1"
    with ensure_kill_all():
        result = th.docker_predict([], tag=tag, network="host", local_port=None)
        assert result == expected_content


def _container_exists(container) -> bool:
    for row in Docker.client().ps():
        if row.id.startswith(container.id):
            return True
    return False


def verify_system_package_installed_on_container(container, pkg: str):
    resp = container.execute(["which", pkg])
    assert resp.strip() == f"/usr/bin/{pkg}"


def verify_system_requirement_not_installed_on_container(container, pkg: str):
    try:
        container.execute(["dpkg", "-l", pkg])
    except DockerException as excp:
        assert "no packages found" in str(excp)


def verify_python_requirement_installed_on_container(container, req: str):
    resp = container.execute(["pip", "show", req])
    assert resp.splitlines()[0].lower() == f"Name: {req}".lower()


def verify_python_requirement_not_installed_on_container(container, req: str):
    try:
        container.execute(["pip", "show", req])
    except DockerException as excp:
        assert "not found" in str(excp)


def verify_environment_variable_on_container(
    container, env_var_name: str, env_var_value: str
):
    resp = container.execute(["env"])
    needle = f"{env_var_name}={env_var_value}"
    assert needle in resp.splitlines()


def _read_readme(readme_correct_path: Path) -> str:
    return readme_correct_path.open().read().replace("\n", "")


def generate_default_config():
    # The test fixture varies with host version.
    python_version = map_local_to_supported_python_version()
    config = {
        "python_version": python_version,
        "resources": {
            "accelerator": None,
            "cpu": "1",
            "memory": "2Gi",
            "use_gpu": False,
        },
    }
    return config


def test_config_verbose(custom_model_truss_dir_with_pre_and_post):
    th = TrussHandle(custom_model_truss_dir_with_pre_and_post)

    new_config = generate_default_config()
    assert new_config == th.spec.config.to_dict(verbose=False)

    th.live_reload()
    new_config["live_reload"] = True
    assert new_config == th.spec.config.to_dict(verbose=False)



import logging
import os
import pathlib
import shutil
import subprocess
import sys

import rich.prompt
from InquirerPy import inquirer

from truss.util import user_config


def _run_upgrade(command: str, console: rich.console.Console) -> bool:
    console.print(f"[bold]Running:[/bold] '{command}'")
    returncode = subprocess.run(command, shell=True).returncode
    if returncode == 0:
        console.print("[green]✅ Upgrade complete. Please re-run your command.[/green]")
        return True
    else:
        console.print(
            f"[bold red]😤 Command failed with exit code {returncode}. "
            "Try upgrading manually.[/bold red]"
        )
        return False


def _is_poetry_venv() -> bool:
    # Typical pattern: poetry puts a `pyvenv.cfg` file inside the venv root
    venv_cfg_path = pathlib.Path(sys.prefix) / "pyvenv.cfg"
    if not venv_cfg_path.exists():
        return False

    text = venv_cfg_path.read_text().lower()
    return "poetry" in text or "virtualenv" in text


def _make_upgrade_command_candidates(latest_version: str) -> list[tuple[str, str]]:
    candidates = []
    if "CONDA_PREFIX" in os.environ:
        candidates.append(("conda", f"conda install truss={latest_version}"))
    if shutil.which("pipx"):
        candidates.append(("pipx", "pipx upgrade truss"))
    if _is_poetry_venv():
        candidates.append(("poetry", f"poetry add truss>={latest_version}"))
    # Pip fallback.
    candidates.append(
        ("pip", f"python -m pip install --upgrade truss=={latest_version}")
    )
    if shutil.which("pipenv"):
        candidates.append(("pipenv", "pipenv update truss"))
    if shutil.which("pdm"):
        candidates.append(("pdm", f"pdm add truss=={latest_version}"))
    if shutil.which("hatch"):
        candidates.append(("hatch", f"hatch dep add truss@{latest_version}"))
    if shutil.which("uv"):
        candidates.append(("uv", f"uv pip install --upgrade truss=={latest_version}"))
    if shutil.which("rye"):
        candidates.append(("rye", f"rye add truss@{latest_version}"))
    return candidates


def upgrade_dialogue(current_version: str, console: rich.console.Console) -> None:
    settings = user_config.settings
    update_info = user_config.state.should_upgrade(current_version)
    latest_version = str(update_info.latest_version)
    logging.debug(f"Truss package update info: {update_info}")
    if not update_info.upgrade_recommended:
        return

    if auto_upgrade_command_template := settings.auto_upgrade_command_template:
        console.print(
            f"[bold yellow]🪄 Automatically upgrading truss to '{latest_version}'.[/bold yellow]"
        )
        command = auto_upgrade_command_template.format(
            version=update_info.latest_version
        )
        if _run_upgrade(command, console):
            sys.exit(0)
        else:
            console.print(
                f"[bold]🖊️  You can edit or remove 'auto_upgrade_command_template' in '{settings.path()}'[/bold]"
            )
            sys.exit(1)

    console.print(
        f"[bold yellow]⬆️  Please upgrade truss. {update_info.reason} → new version "
        f"✨'{latest_version}'✨.[/bold yellow]"
    )

    candidates = _make_upgrade_command_candidates(latest_version)
    do_nothing_cmd = "Do nothing."
    candidates.append(("🫥", do_nothing_cmd))
    options = [f"[{env}] {cmd}" for env, cmd in candidates]

    selection = inquirer.select(
        message="Pick a command for upgrading (you can edit before running):",
        choices=options,
        default=options[0],
    ).execute()

    selected_cmd = next(cmd for label, cmd in candidates if f"[{label}]" in selection)
    if selected_cmd == do_nothing_cmd:
        return

    edited_cmd = inquirer.text(
        message="🖊️  Optionally edit:", default=selected_cmd
    ).execute()

    if inquirer.confirm(
        message="▶️  Run command in this shell?", default=True
    ).execute():
        if _run_upgrade(edited_cmd, console):
            template = edited_cmd.replace(f"{latest_version}", "{version}")
            if inquirer.confirm(
                message="💾 We can remember your choice and automatically upgrade next "
                f"time, using:\n  '{template}' (and filling in a version).\n  You can edit or "
                f"remove this command in '{settings.path()}'.\n  Would you like to enable "
                "automatic upgrades?",
                default=True,
            ).execute():
                settings.auto_upgrade_command_template = template
            sys.exit(0)
        else:
            sys.exit(1)
